/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 *
 * Filename:    yp2.c 
 *
 * Date:        September 1994
 *
 * Purpose:     This file implements the yellow pages module.
 *
 * Author:      Frederic P. H. van Hengstum, Vincent Berkhout.
 *
 * Remarks: 	This file implements the transport level addressing,
 *				accomplished by the tuple (taddress, tdomain) information.
 *
 *
 *							OUT	      Rq Rs
 *						               ^ ^    IPM
 *						----|-|--------|-|----
 *						    v v	              LPM
 *						   Rq Rs	   IN
 *
 * -----------------------------------------------------------------------
 */
#include <yp2.h>




/* 
 * -------------------------------------------------------------------
 *
 * Function:    yp2OutRq
 *
 * Arguments:   id, request identifier (IN/OUT)
 *				type, pdu type (IN)
 *				ta, transport address (IN) 
 *				qos, priviledged and/or authentication (IN)
 *				vi viewindex (IN), le localentity (IN), lt localtime (IN)
 *				src, sourceparty (OUT)
 *				dst, destinationparty (OUT)
 *				ctx, context (OUT)
 *
 * Returns:     OK on success, otherwise ERROR.
 *
 * Purpose:     This function maps all input parameters upon an 
 *				<src,dst,ctx> triple.
 *
 * Remarks:		Aditionally, the UT-APPL ID is remapped upon UT-SPM ID
 *				and the application id and pdu type is stored.	
 *
 * -------------------------------------------------------------------
 */
	STATUS
yp2OutRq( 
    AsnInt *id, AsnInt type, AsnOcts *ta, AsnInt qos, 
	AsnInt vi, AsnOcts *le, AsnOid *lt, 
	AsnInt *src, AsnInt *dst, AsnInt *ctx )
{
	static AsnInt init = 0 ;
	static AsnOcts *community ;
	AsnInt i, version = SNMPVERSION2 ;
	struct timeval tp ;

	if( ! init ) {
		community = pNewOct( 32 ) ;
		for( i = 0; i < YP2OUTLISTMAX; i++ )
		{
			yp2outlist[i].id = NULL ;
			yp2outlist[i].type = 0 ;
			yp2outlist[i].timestamp = 0 ;
		}
		init = 1 ;
	}
	for( i = 0; i < YP2OUTLISTMAX; i++ )
		if( ! yp2outlist[i].id ) {
			if( ypMap( type, qos, lt, ta, le, src, dst, ctx,
					   &version, community ) != OK )
				return ERROR ;
			yp2outlist[i].id = *id ;
			yp2outlist[i].type = type ;
			if( gettimeofday( &tp, NULL ) == -1 ) return ERROR ;
			yp2outlist[i].timestamp = tp.tv_sec ;
			*id = i ;
			return OK ;	
		}
	return ERROR ;
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:    yp2OutRs
 *
 * Arguments:   id, request identifier (IN)
 *				src, sourceparty (OUT)
 *				dst, destinationparty (OUT) 
 *				ctx, context (OUT)
 *
 * Returns:     OK on success, otherwise ERROR.
 *
 * Purpose:     This function maps all input parameters upon an 
 *				<src,dst,ctx> triple.
 *
 * Remarks:		The stored <src,dst,ctx> triple, referenced by id, is 	
 *				retained. The UT-SPM ID is mapped upon REQUEST ID.
 *
 * -------------------------------------------------------------------
 */
	STATUS
yp2OutRs( 	
	AsnInt *id, AsnInt *src, AsnInt *dst, AsnInt *ctx )
{
	AsnInt i ;
	struct timeval tp ;

	if( gettimeofday( &tp, NULL ) == -1 ) return ERROR ;
	for( i = 0; i < YP2INLISTMAX; i++ )
		if( yp2inlist[i].timestamp + YP2INLISTTIMEOUT < tp.tv_sec )
		{
            yp2inlist[i].id = 0 ;
            yp2inlist[i].src = NULL ;
            yp2inlist[i].dst = NULL ;
            yp2inlist[i].ctx = NULL ;
            yp2inlist[i].timestamp = 0 ;
		}
	if( yp2inlist[*id].id ) {
		i = *id ;
		*src = yp2inlist[i].src ;
		*dst = yp2inlist[i].dst ;
		*ctx = yp2inlist[i].ctx ;
		*id = yp2inlist[i].id ;

        yp2inlist[i].id = 0 ;
        yp2inlist[i].src = NULL ;
        yp2inlist[i].dst = NULL ;
        yp2inlist[i].ctx = NULL ;
        yp2inlist[i].timestamp = 0 ;
	
		return( OK ) ;
	}
	return( ERROR ) ;
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:    yp2InRq
 *
 * Arguments:   id, request identifier (IN/OUT)
 *				src, sourceparty (IN)
 *				dst, destinationparty (IN) 
 *				ctx, context (IN)
 *
 * Returns:     OK on success, otherwise ERROR.
 *
 * Purpose:     This function registers <src,dst,ctx> triple from an 
 *				incoming message. 
 *
 * Remarks:		The REQUEST ID is remapped upon UT-SPM ID
 *
 * -------------------------------------------------------------------
 */
	STATUS
yp2InRq( 
	AsnInt *id, AsnInt src, AsnInt dst, AsnInt ctx )
{
	static AsnInt init = 0 ;
	AsnInt i ;
	struct timeval tp ;

	if( ! init ) {
		for( i = 0; i < YP2INLISTMAX; i++ )
		{
			yp2inlist[i].id = 0 ;
            yp2inlist[i].src = NULL ;
            yp2inlist[i].dst = NULL ;
            yp2inlist[i].ctx = NULL ;
            yp2inlist[i].timestamp = 0 ;
		}
		init = 1 ;
	}

	for( i = 0; i < YP2INLISTMAX; i++ )
		if( ! yp2inlist[i].id ) {
			yp2inlist[i].id = *id ;
			yp2inlist[i].src = src ;
			yp2inlist[i].dst = dst ;
			yp2inlist[i].ctx = ctx ;
			if( gettimeofday( &tp, NULL ) == -1 ) return ERROR ;
			yp2inlist[i].timestamp = tp.tv_sec ;
			*id = i ;
			return OK ;	
		}
	return ERROR ;
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:    yp2InRs
 *
 * Arguments:   id, request identifier (IN/OUT)
 *				type, pdu type (OUT)
 *
 * Returns:     OK on success, otherwise ERROR.
 *
 * Purpose:     This function retains the original request pdutype .
 *
 * Remarks:		Additionally, the UT-SPM ID is remapped upon 
 *				UT-APPL ID.
 *
 * -------------------------------------------------------------------
 */
	STATUS
yp2InRs( 
	AsnInt *id, AsnInt *type )
{
	AsnInt i ;
	struct timeval tp ;

	if( gettimeofday( &tp, NULL ) == -1 ) return ERROR ;
	for( i = 0; i < YP2OUTLISTMAX; i++ )
		if( yp2outlist[i].timestamp + YP2OUTLISTTIMEOUT < tp.tv_sec )
		{
			yp2outlist[i].id = yp2outlist[i].type = yp2outlist[i].timestamp = 0 ;
		}
	if( yp2outlist[*id].id ) 
	{
		i = *id ;
		*id = yp2outlist[i].id ;
		*type = yp2outlist[i].type ;
		yp2outlist[i].id = yp2outlist[i].type = yp2outlist[i].timestamp = 0 ;
		return OK ;
	}
	return ERROR ;
}



/* -----------------------------------------------------------------------
 */
