/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 *
 * Filename:    perMISC.c
 *
 * Date:        May 1994.
 *
 * Purpose:     Interface to PER modules.
 *
 * Author(s):   Rene Oude Vrielink
 *
 * -----------------------------------------------------------------------
 */

/* --------------------------------------------------
 * Function:	PerDecLenDeter.
 * Arguments:	length, UAsnInt (OUT)
 *		b, ....,  (IN)
 *		choice, AsnInt (IN)
 * Purpose:	Constuct a length from the length determinant
 *		field. What follows will be an octet-aligned bit field
 *		(for SEQUENCE OF, OCTET STRING, BIT STRING)
 *		(see 9.8 PER-CD)
 * Remarks:	The choice is between SEQUENCE OF,
 *		OCTET STRING, BIT STRING (not yet)
 * ----------------------------------------------------
 */
	STATUS
PerDecLenDeter(UAsnInt *length, BUF_TYPE b, AsnInt choice)
{
 unsigned long int 	retVal;
 unsigned long int	byte;
 unsigned long int	largeVal;

#define ONE_MASK 	0x80  /* bit 8 is set to one */
#define TWO_MASK	0xc0  /* bit 8 & 7 are set to one */

 byte = (unsigned long int) PerGetByte( b );

 if ((byte & ONE_MASK) != ONE_MASK) /* lenght is smaller than 128 */
	*length = byte;
 if ((byte & TWO_MASK) != TWO_MASK) /* lenght is smaller than 16K */
 {
	/* 
	 * bits 7 to 1 will from original byte will
	 * be copied into retVal
	 *
	 */ 
	retVal = (byte & (~ONE_MASK));
	retVal <<= 8; /* shift retVal 1 byte to left */
	byte = PerGetByte(b);
	retVal |= byte; /* copy next byte into retVal */
	*length = retVal; 
 }
 else   /* length is equal or bigger than 16K */
 {
	/* 
	 * bits 6 to 1 will from original byte will
	 * be copied into retVal
	 *
	 */ 
	retVal = (byte & (~TWO_MASK));
	retVal <<= 8; /* shift retVal 1 byte to left */
	largeVal = retVal * 16384; /* multiply the retVal with 16K see 9.8.12
				    * PER Committee Draft
				    */
	retVal = 0;
	PerDecLenDeter(&retVal,b,choice);
	*length = largeVal + retVal;
 }
 return( OK );
}

/* --------------------------------------------------
 * Function:	PerEncLenDeter.
 * Arguments:	length, UAsnInt (IN)
 *		b, ....,  (IN/OUT)
 *		choice, AsnInt (IN)
 * Purpose:	Constuct a length determinant field
 *		to be prepended to a field-list. Outside
 *		this function the determinant field has
 *		to be inserted into the field-list.
 *		(see 9.8 PER-CD)
 * Remarks:	The choice is between SEQUENCE OF,
 *		OCTET STRING, BIT STRING (not yet)
 * ----------------------------------------------------
 */
	STATUS
PerEncLenDeter(UAsnInt *length, BUF_TYPE b, AsnInt choice)
{
  unsigned long int	mcount;


  if (length < 128) /* 
  {
    PerPutByte(b,length);/* set LSB byte into buf */
    /* if length is smaller than 128 then MSbit is always set!*/
  }
  else
  {
    if (length < 16384)
    {
      PerPutByte(b,length);
      length >>= 8; /* shift length 1 byte to the right */
      length &= ONE_MASK;
      /*
       * bit 8 of the LSByte of length is set to one 
       * bit 7 of the LSByte of length is set to zero
       *
       */
      PerPutByte(b,length);
    }
    else 
    {
      if (length >= 16384) /* more than 16 K */ 
      {
	mcount = length/16384;
	if (mcount > 4)
	{
	  mcount = 4;
	}
	lenght -= (mcount*16384);

	PerEncLenDeter(length,b,choice);
	
	mcount &= TWO_MASK;
        /*
         * bit 8 of the LSByte of mcount is set to one 
         * bit 7 of the LSByte of mcount is set to one 
         *
         */

	PerPutByte(b,mcount); 
	
       } /* end if */
     } /* end else if */
  } /* end else if */
  return( OK );
}/* PerLenDeter */



/* ----------------------------------------------------------------------
 * Function:	PerEncInt
 * Arguments:	intvalue, AsnInt ,(IN)
 *		b, ...., (OUT) (unsigned char **)	
 * Purpose:	Encodes an unsigned integer into PER positive
 *		binary intger with as few octets as possible
 * --------------------------------------------------------------
 */

	AsnLen	
PerEncInt(AsnInt *intvalue,BUF_TYPE  b)
{
  int             len;
  int             i;
  unsigned long int   mask;
  unsigned long int   dataCpy;

#define INT_MASK (0xff00 << ((sizeof(AsnInt) - 2) * 8))
#define SIGN_MASK (0x8000 << ((sizeof(AsnInt) - 2) * 8))

  dataCpy = *intvalue;

  mask = INT_MASK; /* the mask is used to check every first 8 bits
		    * this way you be sure to use as few octets 
		    * as possible.
		    */

  if ((dataCpy & SIGN_MASK) == 0)   /* this if statment switches the MSBit of dataCpy */
       dataCpy |= SIGN_MASK
  else
       dataCpy &= ~(SIGN_MASK)

  for (len = sizeof(AsnInt) ; len > 1 ; --len)
   {
 	if ( (dataCpy & mask) == 0)
	      mask >>= 8;
	else
      	break;
   }

   /*
    * write the PER integer
    */
  for (i = 0; i < len; i++)
  {
   	PerPutByte(b ,dataCpy);
	dataCpy >>= 8;
  }

  return(len); /* len is total number of octets needed */
} /* PerEncInt */

/* ----------------------------------------------------------------------
 * Function:	PerEncUInt
 * Arguments:	uintvalue, UAsnInt ,(IN)
 *		b, ....., (OUT)	
 * Returns:	length of the encoding
 * Purpose:	Encodes an unsigned integer into PER positive
 *		binary intger with as few octets as possible
 * --------------------------------------------------------------
 */

	AsnLen	
PerEncUInt(UAsnInt *uintvalue,BUF_TYPE b)
{
  int             len;
  int             retLen;
  int             i;
  unsigned long int   mask;
  unsigned long int   dataCpy;

  dataCpy = *uintvalue;

  /*
   * calculate encoded length of the integer
   */

  mask = INT_MASK;
  for (len = sizeof(UAsnInt) ; len > 1 ; --len)
   {
 	if ( (dataCpy & mask) == 0)
	      mask >>= 8;
	else
      	break;
   }
  
  for (i = 0; i < len; i++)
  {
   	PerPutByte(b,dataCpy);
	dataCpy >>= 8;
  }
  return(len); /* len is total number of octets needed */
}
	
/* ----------------------------------------------------------------
 * Function:	PerDecUInt
 * Arguments:	len ,.... , (IN)
 * 		b,	....,(IN)
 *		uintvalue, ..... , (OUT)
 * Returns:	ERROR or OK
 * Purpose:	Decodes a positive binary integer in PER format
 *		into an unsigned long integer. 
 * Remarks:	The lenght of the PER format is len (in octets)	
 * -----------------------------------------------------------------
 */
	STATUS
PerDecUInt(UAsnInt *uintvalue,BUF_TYPE b, UAsnInt len)
{
  unsigned long int	retVal;
  int             	i;

  
  for (i = 0; i < len; i++)
   	retVal = (retVal << 8) | (unsigned long int) PerGetByte(b);
  
  *uintvalue = (UAsnInt) retVal;
  return( OK ); 
}
/* ----------------------------------------------------------------
 * Function:	PerDecInt
 * Arguments:	intvalue, ..... , (OUT)
 * 		b,	....,(IN)
 *		len, ...., (IN)
 * Returns:	ERROR or OK
 * Purpose:	Decodes a positive binary integer in PER format
 *		into an signed long integer.
 * Remarks:	The lenght of the PER format is len (in octets)	
 * -----------------------------------------------------------------
 */
	STATUS
PerDecInt(AsnInt *intvalue,BUF_TYPE b, UAsnInt len)
{
  unsigned long int	retVal;
  int             	i;


  for (i = 0; i < len; i++)
   	retVal = (retVal << 8) | (unsigned long int) PerGetByte(b);
 
  /* this if statment switches the MSBit of dataCpy 
   * this way you change a PER integer to signed integer
   * e.g. zero becomes -(2^31) 
   */
  if ((retVal & SIGN_MASK) == 0)   
       retVal |= SIGN_MASK
  else
       retVal &= ~(SIGN_MASK)
 
  *intvalue = (AsnInt) retvalue;

  return( OK ); 
}
