/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 *
 * Filename:    spmUnwrap.c
 *
 * Date:        May 1994
 *
 * Purpose:     Unwrapping and processing an incoming message.
 *
 * Author(s):   Vincent Berkhout, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <spmReceive.h>


/* 
 * -------------------------------------------------------------------
 *
 * Function:	SpmUnwrapInit	
 *
 * Arguments:	none 
 *
 * Returns:		STATUS	which is OK on succes
 *
 * Purpose:		assign the value of some global "constants"
 *
 * Remark:		none
 *
 * -------------------------------------------------------------------
 */
	STATUS
unwrapInit ()
{
    return OK;
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:	SpmUnwrapV1Msg	
 *
 * Arguments:	msg, 		received message (IN) 
 *		tag, 		type of pdu (OUT)
 *		errStat, 	errorStatus (OUT).
 *		errInd,  	errorIndex (OUT).
 *		varBinds, 	variable bindings (OUT).
 * 		version,	snmp version number (OUT).
 * 		community, 	community string (OUT).
 *		enterprise, 	(OUT).
 * 		agent_addr, 	(OUT).
 * 		generic_trap, 	(OUT).
 * 		specific_trap,	(OUT).
 * 		time_stamp	(OUT). 
 *
 * Returns:	STATUS	which is OK on succes
 *
 * Purpose:	The SNMP v1 message that is received from the transportlayer
 *		can now be processed, according to rfc1157 and rfc1445 
 *
 * Remarks:	none
 *
 * -------------------------------------------------------------------
 */
	STATUS
SpmUnwrapV1Msg( AsnOcts *msg, PDUsChoiceId *tag, AsnInt *reqId, 
		AsnInt *errStat, AsnInt *errInd, VarBindList **varBinds,
		AsnInt *version, AsnOcts *community, AsnOcts *enterprise, 
		AsnOcts *agent_addr, AsnInt *generic_trap, 
		AsnInt *specific_trap, UAsnInt *time_stamp)
{
	AsnOcts		*pdu=NULL;

	/*
	 *	RFC 1445, par 3.2, step 1, looking at the first octet
	 */ 
	(void) MibIncrementStat( snmpStatsPackets ) ;
	(void) MibIncrementStat( snmpStats30Something );
 
	if (SNMPVERSION1 == FALSE)
		return FAILED;

	if ( BerDecodeV1Message( msg, version, &community, tag, &pdu) != OK) {
		(void) MibIncrementStat( snmpStatsEncodingErrors ); /*?? is this needed */
		return FAILED;
	}

	if ( BerDecodeV1Pdu( pdu, *tag, reqId, errStat, errInd, varBinds, &enterprise, 
		&agent_addr, generic_trap, specific_trap, time_stamp) != OK ) {
		(void) MibIncrementStat( snmpStatsEncodingErrors ); /*?? is this needed */
		return FAILED;
	} 

	return OK ;
} 


/* 
 * -------------------------------------------------------------------
 *
 * Function:	SpmUnwrapV2Msg	
 *
 * Arguments:	msg, 		received message (IN) 
 *		src,		source party identity (OUT).
 *		dst,		destination party identity (OUT).
 *		ctx,		context identity (OUT).
 *		reqId, 		request Identifier (OUT).
 *		tag, 		type of pdu (OUT)
 *		errStat, 	(OUT).
 *		errInd,  	(OUT).
 *		varBinds 	(OUT).
 *
 * Returns:	STATUS	which is OK on succes
 *
 * Purpose:	The message that is received from the transportlayer
 *		can now be processed, according to RFC1445 (and 1446) 
 *
 * Remarks:	none
 *
 * -------------------------------------------------------------------
 */
	STATUS
SpmUnwrapV2Msg( AsnOcts *msg, UAsnInt *tsrc, UAsnInt *tdst, UAsnInt *tctx, 
	   PDUsChoiceId *tag, AsnInt *reqId, AsnInt *errStat, 
	   AsnInt *errInd, VarBindList **varBinds )
{
	AsnOcts		*privData=NULL, *authDigest=NULL, 
			*authData=NULL, *pdu=NULL;
	AsnInt		acl=0, view=0, srcLocal=0, dstLocal=0, contextLocal=0,
			authSrcTimestamp=0, authDstTimestamp=0, authDigestOffset=0;
	AsnOid		*privDst=NULL, *src=NULL, *dst=NULL, *ctx=NULL;
	UAsnInt		dstIndex, srcIndex, ctxIndex ;
	char		Tag ;

	/*
	 *	RFC 1445, par 3.2, step 1
	 * 	SnmpPrivMsg ::= [1] IMPLICIT SEQUENCE {
	 *		privDst
	 *			OBJECT IDENTIFIER,
	 *		privData
	 *			[1] IMPLICIT OCTET STRING
	 *	}				
	 */ 
	(void) MibIncrementStat( snmpStatsPackets ) ;
 
 	if( BerDecodePrivMsg( msg, &privDst, &privData ) != OK ) {
		Error( "encoding error [1]" ) ;
		MibIncrementStat( snmpStatsEncodingErrors );
		return FAILED ;
	}

 	if( GetPartyIndex( privDst, &dstIndex) != OK )	
	{
		Error( "privDestination unknown [1]" ) ;
		return FAILED ;
	}


	/*	RFC 1445, par 3.2, step 2 (MibGetPartyLocal) and 3 
	 */ 
	if( ( MibGetPartyLocal( dstIndex, (TruthValue *) &dstLocal ) != OK ) || 
	    ( dstLocal != true ) )	
	{
		Error( "destination party unknown [2]" ) ;
		MibIncrementStat( snmpStatsUnknownDstParties );
		return FAILED ;
	}

	/*	RFC 1445, par 3.2, step 4 (DecryptPrivData) and 5 
	 */ 
	if( UnwrapDecryptPrivData( dstIndex, privData ) != OK )
	{
		Error( "decryption error [4]" ) ;
		MibIncrementStat( snmpStatsEncodingErrors );
 		return FAILED ;
	}

	if( BerDecodeAuthMsg(privData, &authDigest, &authDstTimestamp, 
		&authSrcTimestamp, &authDigestOffset, &authData) != OK )
	{
		Error( "authMsg decoding error [5]" ) ;
		MibIncrementStat( snmpStatsEncodingErrors );
		MibIncrementStat( snmpStatsBadAuths );
		return FAILED ;
	}

	if( BerDecodeCommMsg( authData, &src, &dst, &ctx, &pdu, tag ) != OK )
	{
		Error( "commMsg decoding error [5]" ) ;
		MibIncrementStat( snmpStatsEncodingErrors );
		return FAILED ;
	}



	/*
	 *	RFC 1445, par 3.2, step 6 : privDstOid is obsolete
	 */
	if( CmpOID( dst, privDst ) == FALSE ) 
	{
		Error( "destination party mismatches [6]" ) ;
		MibIncrementStat( snmpStatsDstPartyMismatches );
		return FAILED ;
	}
	/*
	 *	fill in the indices for the cache entries
	 */ 
	if( ( GetPartyIndex( src, &srcIndex) != OK ) ||
	    ( GetPartyIndex( dst, &dstIndex) != OK ) ||
	    ( GetContextIndex( ctx, &ctxIndex) != OK ) ) {

		Error( "party adresses not found [6]" ) ;
		return FAILED ;
	}

		 
	/*
	 *	RFC 1445, par 3.2, step 7 (GetParty) and 8 
	if( ( MibGetPartyLocal( srcIndex, &srcLocal ) != OK ) ||
	    ( srcLocal == true ) )
	{
		Error( "source party unknown [7]" ) ;
		MibIncrementStat( snmpStatsUnknownSrcParties );
		return FAILED ;
	}
	 */

	/*
	 *	RFC 1445, par 3.2, step 9 (AuthenticateSnmpAuthMsg) and 10 
	 */ 
	if( AuthenticateSnmpAuthMsg( srcIndex, dstIndex, authSrcTimestamp, 
		authDstTimestamp, authDigest, authDigestOffset,
		privData ) == UNAUTH ) 
	{
		/*
		if ( MibGetTrap( snmpV2EnableAuthenTrap, 0 ) )
		{		zie rfc 1450 page 15 
			generate trap ;
		}
		*/
		Error( "unauthenticated message [10]" ) ;
		return FAILED ;
	} 		

	/*
	 *	RFC 1445, par 3.2, step 11, 12 (MibGetContext) and 13
	 */ 
	if( MibGetContextLocal( ctxIndex, (TruthValue *) &contextLocal ) != OK ) {
		Error( "context unknown [13]" ) ;
		MibIncrementStat( snmpStatsUnknownContexts );
		return FAILED ;
	}


	/*
	 *	RFC 1445, par 3.2, step 14
	 */ 
	if( MibGetAcl( &dstIndex, &srcIndex, &ctxIndex, &acl ) != OK ) {
		Error( "acl entry not found [14]" ) ;
		MibIncrementStat( snmpStatsUnknownContexts );
		return FAILED ;
	}

	/*
	 *	RFC 1445, par 3.2, step 15
	 */ 
	if( BerDecodeV2Pdu(pdu, tag, reqId, errStat, errInd, varBinds) != OK ) {
		Error( "wrong pdu decoding [15]" ) ;
		MibIncrementStat( snmpStatsUnknownContexts );
		return FAILED ;
	}	

	Tag = 1 << ( 0x07 & *tag) ; 

	if( ( Tag & AGENTMASK) && (contextLocal != true)  )
	{
		Error( "context unknown [15]" ) ;
		MibIncrementStat( snmpStatsUnknownContexts );
		return FAILED ;
	}

	/*
	 *	RFC 1445, par 3.2, step 16
	 */ 
	if( (Tag & MANAGERMASK) && !(Tag & acl) ) 
	{
		Error( "(pdu) operation not supported [16]" ) ;
		MibIncrementStat( snmpStatsBadOperations );
		return FAILED ;
	}

	/*
	 *	RFC 1445, par 3.2, step 17
	 */ 
	if( !(Tag & acl) ) {
		Error( "operation not supported" ) ;
		return FAILED ;
	}

	if( MibGetContextViewIndex( srcIndex, &view ) != OK ) {
		Error( "context unknown [17]" ) ;
		MibIncrementStat( snmpStatsUnknownContexts );
		return FAILED ; 
	}

	if( view ) {
		/*
		 *	RFC 1445, par 3.2, step 18
		 */ 
	} else { 
		/* 
		 *	RFC 1445, par 3.2, step 19
		 */
	}

	*tsrc=srcIndex; *tdst=dstIndex; *tctx=ctxIndex;

	return OK ;

}	/* end of function spmUnwrap */



/* 
 * -------------------------------------------------------------------
 *
 * Function:	UnwrapDecryptPrivData
 *
 * Arguments:	OID		*privDstOid	(in)
 * 		OCTETSTR	*privData	(in/out) 
 *
 * Returns:	STATUS	which is OK on succes
 *
 * Purpose:	select the appropriate privacy protocol and call the 
 *		associated algorithm
 *
 * Remarks:	This function implements RFC1445, par 3.2, step 4 and
 *		RFC1446, par 4.2
 *
 * -------------------------------------------------------------------
 */
	STATUS	
UnwrapDecryptPrivData( UAsnInt privDstIndex, AsnOcts *privData )
{
	AsnOid	*protocol ;
	AsnOcts	*privPrivateKey ;

	protocol = NewOid(9) ;

	if ( MibGetPartyPrivProtocol( privDstIndex, protocol ) != OK )	{
		(void) FreeOid(protocol) ;
		return FAILED ;
	}
 
	if( CmpOID( protocol, &noPriv ) == TRUE )	{
		(void) FreeOid(protocol) ;
		return OK ;
	}

	if( CmpOID( protocol, &desPrivProtocol ) == TRUE ) {
		privPrivateKey = NewOct(16) ;
 		if ( MibGetPartyPrivPrivate( privDstIndex, privPrivateKey ) == OK ) 
			if( DesDecrypt( privPrivateKey, privData ) == OK ) {
				(void) FreeOid( protocol ) ;
				(void) FreeOid( privPrivateKey ) ;
				return OK ;
			}
	}

	(void) FreeOid( protocol ) ;
	(void) FreeOid( privPrivateKey ) ;

	return FAILED ;		
}



/* 
 * -------------------------------------------------------------------
 *
 * Function:	AuthenticateSnmpAuthMsg
 *
 * Arguments:	OID		*dstOid		(in)
 *		OID		*srcOid		(in)
 * 		OCTETSTR	*authInfo 	(in)
 * 		OCTETSTR	*authData	(in)
 *
 * Returns:	STATUS:	AUTH when the authentication succeeded,
 * 			UNAUTH otherwise
 *
 * Purpose:	Athenticate the authData with the authInfo and keys provided
 *		by the receiving and sending parties
 *
 * Remarks:	
 *
 * -------------------------------------------------------------------
 */
	STATUS
AuthenticateSnmpAuthMsg( UAsnInt srcIndex, UAsnInt dstIndex, AsnInt srcTimestamp, 
			 AsnInt dstTimestamp, AsnOcts *digest, AsnInt digestOffset, 
			 AsnOcts *privData )
{
	AsnOid	*protocol ;
	AsnOcts	*srcPrivKey=NULL, *newDigest=NULL;
	AsnInt	srcClock=0, dstClock=0, srcLifetime=0, acl=0;
	struct 	timeval	now ;
	UAsnInt ctxIndex=0;

	protocol = NewOid(16) ;
	srcPrivKey = NewOct(16) ;
	newDigest = NewOct(16);
	srcPrivKey->octetLen = newDigest->octetLen = 16;

	/*
	 *	RFC 1445, par 3.2 step 9, 
	 */ 

	if ( MibGetPartyAuthProtocol( dstIndex, protocol ) != OK )
		return UNAUTH;
	
	if ( CmpOID( protocol, &noAuth ) == TRUE ) 
		return AUTH ;

	if ( CmpOID( protocol, &v2md5AuthProtocol ) == TRUE ) { 
		/*
		 * 	RFC 1446, par 3.1, step 1 -> decode ASN1 authInfo component
		 * 
		 *	AuthInformation ::= [2] IMPLICIT SEQUENCE {
		 * 	authDigest
		 * 		OCTET STRING
		 * 	authDstTimestamp
		 * 		UInteger32
		 * 	authSrcTimestamp
		 * 		UInteger32
		 * 	}
		 */  
		/*
		 *	RFC 1446, par 3.1, step 2: retreive authentication info
		 */
		if ( 	( MibGetPartyAuthClock( srcIndex, &srcClock ) != OK ) ||
			( MibGetPartyAuthPrivate( srcIndex, srcPrivKey ) != OK ) ||
			( MibGetPartyAuthLifetime( srcIndex, &srcLifetime ) != OK ) ) 
			return UNAUTH ;


		/*
		 *	RFC 1446, par 3.1, step 3 : check lifetime
		 */
		(void) gettimeofday ( &now , (struct timeval * ) NULL ) ;
		if( ((srcTimestamp=now.tv_sec) + srcLifetime) < srcClock ) {
			(void) MibIncrementStat( snmpStatsNotInLifetimes );
			return UNAUTH ;	
		}


		/*
		 *	RFC 1446, par 3.1, step 5: determine the new digest 
		 */
		(void) memcpy( (privData->octs) + digestOffset, 
			       srcPrivKey->octs, srcPrivKey->octetLen );
		MD5( newDigest, privData ) ;
 

		/*
		 *	RFC 1446, par 3.1, step 6: compare digests
		 */
		if( CmpOID( digest, newDigest ) == FALSE  ) { 
			(void) MibIncrementStat( snmpStatsWrongDigestValues );
			return UNAUTH ;	        
		}
			
		/*
		 *	RFC 1446, par 3.1, step 8 : validate clocks
		 */
		if ( MibGetAcl( srcIndex, dstIndex, &ctxIndex, &acl ) == OK  ) {
			if( srcClock < srcTimestamp)
				(void) MibPutPartyAuthClock( srcIndex, srcTimestamp ) ;
			if( MibGetPartyAuthClock( dstIndex, &dstClock ) == OK ) 
				if( dstClock < dstTimestamp)
					(void) MibPutPartyAuthClock( dstIndex, dstTimestamp ) ;
		}
		/*
		 *	RFC 1446, par 3.1, step 7 : the message is authentic!
		 */		
		return AUTH ; 

	} 	/* endif V2MD5 */

	return UNAUTH ;		/* unsupported authentication protocol */ 

} 	/* end of function AuthenticateSnmpAuthMsg */

/*---------------------------- end of file spmUnwrap.c ------------------------*/

