/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 *
 * Filename:    spmWrap.c
 *
 * Date:        May 1994
 *
 * Purpose:     Wraping and encoding a (asn1) message. 
 *
 * Author(s):   Vincent Berkhout, Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <spmSend.h>	/* all includes are the same as one level higher! */

SpmWrapV1Message(AsnOcts *community, AsnOcts *tAddress, PDUsChoiceId tag, 
		AsnInt requestid, AsnInt errorstatus, AsnInt errorindex, 
		VarBindList *variablebindings, AsnOcts **message) 

{
	MessageV1Int version;
	AsnOcts	*agent_addr=NULL, *enterprise=NULL;
	AsnInt generic_trap=0, specific_trap=0;
	UAsnInt time_stamp=0;

	/* 
	 * v1_message ::=
	 * 	version, INTEGER
	 * 	community, OCTETSTRING
	 * 	data, ANY
	 */ 

	version = VERSION_1;

	enterprise = NewOct(128) ;
	agent_addr = NewOct(32) ;
	enterprise->octetLen = agent_addr->octetLen = 0 ;

	if (tag == PDUS_TRAP)	{
	    	if  (VarbindlistToV1Trap( &variablebindings, enterprise,
			&generic_trap, &specific_trap, &time_stamp) != OK) {
			return FAILED; 
		}
	}

	if ( BerEncodeV1Message ( message, version, community, tag, 
		requestid, errorstatus, errorindex, variablebindings,
		enterprise, agent_addr, generic_trap, specific_trap, 
		time_stamp) != OK) {
		return FAILED;
	} 

	return OK ;
}


/* 
 * -------------------------------------------------------------------
 *
 * Function:	spmWrapV2Message
 *
 * Arguments:	srcOid,		identification of originating party (in). 	
 *		dstOid,		identifictaion of receiving party (in). 
 *		cntOid, 	desired SNMPv2 context (in).
 *		tag,		desired management operation (in) 
 * 		requestId, 	the unique requestid provided by the application (in).
 * 		errorStatus,	indication about the possible occured error (in).
 * 		errorIndex, 	index of varbiniding which produced the error (in).
 *		varBinds, 	the variable binding list (in).
 * 		snmpPrivMsg	the wrapped message, ready for sending
 *
 * Returns:	STATUS	which is OK on succes
 *
 * Purpose:	This is the procedure to generate a request/trap or a response
 *		This is done according to RFC1445 par 3.1 and par 3.3
 *
 * Remarks:	Mind that the steps 1 and 7 are different, w.r.t. retreiving used
 * 		components from partymib versus the original request!
 *
 * -------------------------------------------------------------------
 */
	STATUS
SpmWrapV2Message( UAsnInt tsrc, UAsnInt tdst, UAsnInt tctx, AsnInt *tag, AsnInt reqId,
	 AsnInt *errStat, AsnInt *errInd, VarBindList *varBinds, 
	 AsnOcts **snmpPrivMsg )
{
	AsnOcts	*authData=NULL, *privData=NULL, *authInfo=NULL, *snmpAuthMsg=NULL;
	AsnOid *src=NULL, *dst=NULL, *ctx=NULL;

	src=NewOid(256); 
	dst=NewOid(256); 
	ctx=NewOid(256);
	src->octetLen =  dst->octetLen = ctx->octetLen = 0;

	if( ( MibGetPartyID( tsrc, src ) != OK ) ||
	    ( MibGetPartyID( tdst, dst ) != OK ) ||
	    ( MibGetContextID( tctx, ctx ) != OK ) ) 
		return FAILED ;

	/*
	 *	RFC 1445, par 3.1, step 1
	 * 
	 * 	SnmpMgmtCom ::= [2] IMPLICIT SEQUENCE {
	 *		dstParty
	 *			OBJECT IDENTIFIER,
	 *		srcParty
	 *			OBJECT IDENTIFIER,
	 *		context
	 *			OBJECT IDENTIFIER,
	 *		pdu
	 *			PDUs
	 *				
	 */ 
	if( BerEncodeMgmtCom( src, dst, ctx, tag, reqId, errStat, errInd,
			      varBinds,	&authData ) != OK ) 
		return FAILED ;



	/*
	 *	RFC 1445, par 3.2, step 2 and 3 : authenticate message
	 */ 
	if( WrapAuthenticateSnmpAuthMsg( tsrc, tdst, authData, 
				         &snmpAuthMsg ) != OK ) 
		return FAILED;

	/*
	 *	RFC 1445, par 3.2, step 4 and 5 
	 */ 
	if( WrapEncryptPrivData( tdst, snmpAuthMsg ) != OK ) 
		return FAILED ;


	/* 
	 *	SnmpPrivMsg ::= [1] IMPLICIT SEQUENCE {
	 *		privDst (equals dstOid)
	 *			OBJECT IDENTIFIER,
	 *		privData
	 *			[1] IMPLICIT char STRING
	 *	}				
	 */ 
 
  	if( BerEncodePrivMsg( dst, snmpAuthMsg, snmpPrivMsg ) != OK ) 
		return FAILED ;
	
	return OK ;

}	/* end of function spmWrap */




/* 
 * -------------------------------------------------------------------
 *
 * Function:	WrapEncryptPrivData
 *
 * Arguments:	OID		dstOid		(in)
 * 		charSTR	*snmpData	(in/out) 
 *
 * Returns:	STATUS	which is OK on succes
 *
 * Purpose:	Select the appropriate privacy protocol and call the 
 *		associated algorithm to encrypt the the data
 *
 * Remarks:	implemented according to RFC1445, par 3.2, step 4 and
 *		RFC1446, par 4.2
 *
 * -------------------------------------------------------------------
 */
	STATUS	
WrapEncryptPrivData( UAsnInt dst, AsnOcts *privData )
{
	AsnOid *protocol=NULL ;
	AsnOcts	*privPrivateKey=NewOct(128) ;

	protocol = NewOid(128);
	if( MibGetPartyPrivProtocol( dst, protocol ) != OK )
		return FAILED ;

	if( CmpOID( protocol, &noPriv ) == TRUE )
		return OK ;

	if( CmpOID( protocol, &desPrivProtocol ) == TRUE ) {
		if( MibGetPartyPrivPrivate( dst, privPrivateKey ) == OK )
			if( DesEncrypt( privPrivateKey, privData ) != OK ) 
			{
				Error( "Privicy failure" ) ;
				return( FAILED ) ;
			} else {
				return( OK ) ;
			}
	} 

	return FAILED ;	
}


/* 
 * -------------------------------------------------------------------
 *
 * Function:	WrapAuthenticateSnmpAuthMsg
 *
 * Arguments:	OID		dstOid		(in)
 *		OID		srcOid		(in)
 * 		charSTR	*authData	(in)
 * 		charSTR	**authMsg	(out)
 *
 * Returns:	STATUS which is OK on succes 
 *
 * Purpose:	Authenticate the authData with the authInfo and keys 
 *		provided by the receiving and sending parties
 *
 * Remarks:	Implemented according to rfc1445, par 3.1 step 2 and 3 
 * 		and rfc1446, par 3.1, steps 1 through 5 
 *
 * -------------------------------------------------------------------
 */
	STATUS
WrapAuthenticateSnmpAuthMsg( UAsnInt src, UAsnInt dst, AsnOcts *authData,
			     AsnOcts **authMsg )
{
	AsnOid *protocol=NewOid(256);
	AsnOcts	*srcPrivateKey=NewOct(256), *digest=NewOct(256);
	AsnInt dstTimestamp=0, srcTimestamp=0, 
	       digestOffset=0, srcClock=0, dstClock=0 ;
	struct timeval now ;


	if( MibGetPartyAuthProtocol( dst, protocol ) != OK ) {
		FreeOid(protocol);
		return FAILED ;
	}



	if( CmpOID( protocol, &noAuth ) == TRUE )
		return( BerEncodeAuthMsg( digest, dstTimestamp, srcTimestamp, 
					  &digestOffset, authData, authMsg ) );



	if( CmpOID( protocol, &v2md5AuthProtocol ) == TRUE ) { 
		/*
		 *	RFC 1446, par 3.1, step 1 : retreive authentication info
		 */

		srcPrivateKey = NewOct(256);
		if (	( MibGetPartyAuthClock( dst, &dstClock ) != OK ) ||
			( MibGetPartyAuthClock( src, &srcClock ) != OK ) ||
			( MibGetPartyAuthPrivate( src, srcPrivateKey ) != OK ) ) 
			return FAILED ; 

		/*
		 *	RFC 1446, par 3.1, step 2
		 */
		(void) gettimeofday( &now, (struct timezone * ) 0 ) ;
		srcTimestamp = now.tv_sec ;
		dstTimestamp = dstClock ; 



		/*
		 *	RFC 1446, par 3.1, step 3
		 * 	set digest temporarily to privateKey and encode 
		 *	the message
		 * 
		 *	AuthInformation ::= [2] IMPLICIT SEQUENCE {
		 * 		authDigest
		 * 			char STRING
		 * 		authDstTimestamp
		 * 			UInteger32
		 * 		authSrcTimestamp
		 * 			UInteger32
		 * 	}
		 * 	
		 * 	SnmpAuthMsg ::= [1] IMPLICIT SEQUENCE {
		 * 		authInfo
		 * 			ANY,
		 * 		authData
 		 * 			SnmpMgmtCom
		 * 	} 
		 */  
		if ( BerEncodeAuthMsg( srcPrivateKey, srcTimestamp, dstTimestamp, 
				       &digestOffset, authData, authMsg ) != OK )
		{
			Error( "authentication failure" ) ;
			return FAILED ;
		}

		if ( MD5( digest, *authMsg ) != OK )
		{
			Error( "authentication failure" ) ;
			return FAILED;
		}

		(void) memcpy( (*authMsg)->octs + digestOffset, 
			       digest->octs, digest->octetLen );

		return OK ;
	}

	return FAILED ;	

} 	/* end of function WrapAuthenticateSnmpAuthMsg */

/*---------------------------- end of file spmWrap.c ------------------------*/

