/* -----------------------------------------------------------------------
 *
 * Project:     UT-SNMP
 *
 * Filename:    mibreg.c
 *
 * Date:        September 1994
 *
 * Purpose:     This file implements the core-agent (spm) it`s mib regi
 *				stration from proprietary mib (subtrees) of sub-agents. 
 *
 * Author(s):   Frederic P. H. van Hengstum
 *
 * -----------------------------------------------------------------------
 */

#include <snmpTypes.h> 
#include <UT-SNMP.h>


typedef struct _mibreg {
	AsnOcts *appl ;								   /* name of sub-agent */
	AsnInt id ;									  /* subtree identifier */
	AsnOcts *le ;								         /* localentity */
	AsnOid *lt ;										   /* localtime */
	AsnOid *lb ;								   /* low-bound mibtree */
	AsnOid *ub ;								 /* upper-bound mibtree */
	struct _mibreg *next ;
} MIBREG;									 /* MIB REGistration record */

static MIBREG *MibRegList;				/* the actual registration list */



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mrCheck
 *
 * Arguments:	type, operation type (IN)
 *				var, variable (oid) (IN)
 *				le, localentity (IN)
 *				lt, localtime (IN)
 *				appl, sub-agent name (OUT)
 *
 * Returns:		OK means disjunct, otherwise ERROR.
 *
 * Purpose:		This function checks whether or not an variable is 
 *				included in an arbitrary subtree with respect to the
 *				localentity and localtime.
 *
 * Remarks:		none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
mrCheck( AsnInt type, AsnOid *var, AsnOcts *le, AsnOid *lt, AsnOcts *a )
{
	MIBREG *p = MibRegList ;

	switch( type )
	{
	case UTB_GETIN_TYPE :

		while( p ) 
		{
			if( ( memcmp( le->octs, p->le->octs, le->octetLen ) == 0 ) &&
				( memcmp( lt->octs, p->lt->octs, lt->octetLen ) == 0 ) && 
				( oidcmp( var, p->lb ) >= 0 ) && 
				( oidcmp( var, p->ub ) <= 0 ) )
			{
				memcpy(a->octs,p->appl->octs,a->octetLen=p->appl->octetLen);
				return OK ;
			}
			p = p->next ;
		}
		break ;
 
	case UTB_GETBULKIN_TYPE :
	case UTB_GETBULKRS_TYPE :
	case UTB_GETNEXTIN_TYPE :
	case UTB_GETNEXTRS_TYPE :

		while( p ) 
		{
			if( ( memcmp( le->octs, p->le->octs, le->octetLen ) == 0 ) &&
				( memcmp( lt->octs, p->lt->octs, lt->octetLen ) == 0 ) ) 
			{
				if( oidcmp( var, p->lb ) >= 0 ) 
				{
				  	if( oidcmp( var, p->ub ) < 0 )
					{
						memcpy( a->octs, p->appl->octs,
								a->octetLen = p->appl->octetLen) ;
						return OK ;
					}
                	if( oidcmp( var, p->ub ) == 0 )
					{
						if( p->next ) 
						{
                			memcpy( a->octs, p->next->appl->octs, 
									a->octetLen = p->next->appl->octetLen ) ;
                			return OK ;
						}
						else break ;
            		}
				}
			}
			p = p->next ;
		}
		break ;
	}
	return ERROR ;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mrIntersect
 *
 * Arguments:	lb1, lower-bound subtree 1 (IN)
 *				ub1, upper-bound subtree 1 (IN)
 *				lb2, lower-bound subtree 2 (IN)
 *				ub2, upper-bound subtree 2 (IN)
 *
 * Returns:		OK means disjunct, otherwise ERROR.
 *
 * Purpose:		This function checks whether or not two mib-registrations
 *				are disjunct.
 *
 * Remarks:		none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
mrIntersect( AsnOid *lb1, AsnOid *ub1, AsnOid *lb2, AsnOid *ub2 )
{
	if( oidcmp( lb1, lb2 ) < 0 ) 
		if( oidcmp( ub1, lb2 ) < 0 ) return OK ; 
		else return ERROR ;
	if( oidcmp( lb1, ub2 ) > 0 )
		if( oidcmp( ub1, ub2 ) > 0 ) return OK ; 
		else return ERROR ;
	return ERROR ;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mrInsert
 *
 * Arguments:	n, new subtree registration (IN)
 *
 * Returns:		OK on success, otherwise ERROR.
 *
 * Purpose:		This function inserts an new subtree registration in an
 *				alfanumerical ordered list of subtree registrations.
 *
 * Remarks:		none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
mrInsert( MIBREG *n )
{
	MIBREG *p = NULL, *pp = NULL ;

	for( pp = NULL, p = MibRegList; p; pp = p, p = p->next )
	{
		if( ( n->le->octetLen == p->le->octetLen ) &&
			( n->lt->octetLen == p->lt->octetLen ) )
		{
			if( ( memcmp( n->le->octs, p->le->octs, n->le->octetLen ) == 0 ) 
				&& ( oidcmp( n->lt, p->lt ) == 0 ) )
			{
				if( oidcmp( n->ub, p->lb ) < 0 ) break ;
			}
		}
	}
	if( !pp ) 
	{
		n->next = MibRegList ; MibRegList = n ;
	} else {
		n->next = pp->next ; pp->next = n ;
	}
	return OK ;
}


/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mrRegister
 *
 * Arguments:	appl, sub-agent name (IN)
 *				id, mib subtree identifier (IN)
 *				le, localentity (IN)
 *				lt, localtime (IN)
 *				lb, lower-bound subtree (IN)
 *				ub, upper-bound subtree (IN)
 *
 * Returns:		OK on success, otherwise ERROR
 *
 * Purpose:		This function registers an mib from an sub-agent.
 *
 * Remarks:		none
 *
 * -----------------------------------------------------------------------
 */
	STATUS
mrRegister( AsnOcts *appl, AsnInt id, 
		    AsnOcts *le, AsnOid *lt, AsnOid *lb, AsnOid *ub )
{
	MIBREG *p = MibRegList ;

	while( p ) {
		if( (strcmp(appl->octs, p->appl->octs) == 0) && (id == p->id) )
			return ERROR ;
		if( (strncmp(le->octs, p->le->octs, le->octetLen) != 0) ||
			(strncmp(lt->octs, p->lt->octs, le->octetLen) != 0) ||
			(mrIntersect( lb, ub, p->lb, p->ub ) == OK ) ) p = p->next ;
		else return ERROR ;
	}

	p = (MIBREG *) malloc( sizeof( MIBREG ) ) ;
	p->id = id ;

	p->appl = pNewOct( appl->octetLen ) ;
	memcpy( p->appl->octs, appl->octs, p->appl->octetLen = appl->octetLen ) ; 

	p->le = pNewOct( le->octetLen ) ;
	memcpy( p->le->octs, le->octs, p->le->octetLen = le->octetLen ) ; 

	p->lt = pNewOid( lt->octetLen ) ;
	memcpy( p->lt->octs, lt->octs, p->lt->octetLen = lt->octetLen ) ; 

	p->lb = pNewOid( lb->octetLen ) ;
	memcpy( p->lb->octs, lb->octs, p->lb->octetLen = lb->octetLen ) ; 

	p->ub = pNewOid( ub->octetLen ) ;
	memcpy( p->ub->octs, ub->octs, p->ub->octetLen = ub->octetLen ) ; 

	return mrInsert( p ) ;
}



/* 
 * -----------------------------------------------------------------------
 *
 * Function:	mrUnregister
 *
 * Arguments:	appl, sub-agent name (IN)
 *				id, registration identifier (IN)
 *
 * Returns:		OK on success, otherwise ERROR.
 *
 * Purpose:		This function removes the mib registration from the 
 *				administration.
 *
 * Remarks:		none.
 *
 * -----------------------------------------------------------------------
 */
	STATUS
mrUnRegister( AsnOcts *appl, AsnInt id )
{
	MIBREG *n = NULL, *p = MibRegList ;

	while( p ) {
		if( (strcmp( appl->octs, p->appl->octs ) == 0) && (id == p->id) ) {
			if( n ) n->next = p->next ; else MibRegList = MibRegList->next ;
			Asn1Free( p->appl ) ; Asn1Free( p->lb ) ; Asn1Free( p->ub ) ;
			Asn1Free( p->le ) ; Asn1Free( p->lt ) ; Asn1Free( p ) ;
			return OK ;
		}
		n = p ; p = p->next ;
	}
	return ERROR ;
}



/* -----------------------------------------------------------------------
 */
